///
/// file   : modulator.cpp
/// author : (c) 2001-2003 Bastian Spiegel <bs@tkscript.de>
///
/// changelog:
//        06042002 <fli> wrote interface, constants
//        07042002 <fli> implementations

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#define YAC_BIGSTRING defined
#define YAC_PRINTF defined
#include <yac.h>
#include <tks-list.h>
#include "modulator.h"

extern sUI class_ID_Modulator;
extern YAC_Host *yac_host;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                   ~
//~ TKSModulator                      ~
//~                                   ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
YAC_C(TKSModulator, "Modulator");

TKSModulator::TKSModulator(void) {
    float_array=0;
    initArgs();
}

TKSModulator::~TKSModulator(void) {
    sU8 i;
    for(i=0; i<MOD_NUMARGS; i++)
    {
        if(arg_mod[i])
		{
			delete arg_mod[i];
			arg_mod[i]=0;
		}
	}
}

void TKSModulator::yacGetConstantStringList(YAC_String *_ret) {
    _ret->copy(
"MOD_TYPE_SIN:0 "
"MOD_TYPE_TRI:1 " 
"MOD_TYPE_SAW:2 "
"MOD_TYPE_RECT:3 "
"MOD_TYPE_NOISE:4 "
"MOD_TYPE_1BYX:5 "
"MOD_TYPE_ARRAY:6 "
"MOD_TYPE_SIN01:7 "
"MOD_TYPE_TRI01:8 "
"MOD_TYPE_SAW01:9 "
"MOD_TYPE_RECT01:10 "
"MOD_TYPE_NOISE01:11 "
"MOD_TYPE_1BYX01:12 "

"MOD_ARG_TYPE:0 "
"MOD_ARG_AMP:1 "
"MOD_ARG_OFF:2 "
"MOD_ARG_SPEED:3 "
"MOD_ARG_PULSEWIDTH:4 "
);
}

void TKSModulator::initArgs(void) {
    args[MOD_ARG_TYPE]=MOD_SAW01;
    args[MOD_ARG_SPEED]=0.001f;
    args[MOD_ARG_OFF]=0.0f;
    args[MOD_ARG_AMP]=1.0f;
    args[MOD_ARG_PULSEWIDTH]=0.0f;
    sU8 i;
    for(i=0; i<MOD_NUMARGS; i++)
    {
        arg_mod[i]=0;
        arg_modamount[i]=1.0f;
        arg_modamountmod[i]=0;
    }
}

void TKSModulator::yacOperator(sSI _cmd, YAC_Object *_o, YAC_Value *_r) {
    if(_cmd==YAC_OP_ASSIGN)
    {
        if(YAC_CHK(_o, class_ID_Modulator))
        {
            TKSModulator *o=(TKSModulator*)_o;
            int i;
            for(i=0; i<MOD_NUMARGS; i++)
            {
                args[i]=o->args[i];
            }
        }
    }
}


void TKSModulator::tksinter_setArg(sSI _nr, sF32 _val) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        args[_nr]=_val;
    }
}


void TKSModulator::tksinter_setArgModulation(sSI _nr, sF32 _val) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        arg_modamount[_nr]=_val;
    }
}


void TKSModulator::tksinter_getArg(sSI _nr, YAC_Value *_r) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        YAC_RETF(args[_nr]);
    }
    else
    {
        YAC_RETF(0.0);
    }
}


void TKSModulator::tksinter_setArgModulator(sSI _nr, YAC_Object *_m, sF32 _amount) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        //IFVALIDOBJ(_m)
        if(YAC_CHK(_m, class_ID_Modulator))
        {
            arg_mod[_nr]=(TKSModulator*)_m;
            arg_modamount[_nr]=_amount;
        }
    }
}

void TKSModulator::tksinter_getArgModulator(sSI _nr, YAC_Value *_r) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        YAC_RETO(arg_mod[_nr], 0);
    }
    else
    {
	    YAC_RETO(0, 0);
    }
}

void TKSModulator::tksinter_setArgModulationModulator(sSI _nr, YAC_Object *_m, sF32 _modamountmodamount) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        //IFVALIDOBJ(_m)
        if(YAC_CHK(_m, class_ID_Modulator))
        {
            //arg_modamount[_nr]=_amount;
            arg_modamountmodamount[_nr]=_modamountmodamount;
            arg_modamountmod[_nr]=(TKSModulator*)_m;
        }
    }
}

void TKSModulator::tksinter_getArgModulationModulator(sSI _nr, YAC_Value *_r) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
        YAC_RETO(arg_modamountmod[_nr], 0);
    }
    else
    {
        YAC_RETO(0, 0);
    }
}

void TKSModulator::tksinter_getArgModulationAmount(sSI _nr, YAC_Value *_r) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
		YAC_RETF(arg_modamount[_nr]);
    }
    else
    {
        YAC_RETF(0.0f);
    }
}

void TKSModulator::tksinter_setArgModulationAmount(sSI _nr, sF32 _amount) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
		arg_modamount[_nr]=_amount;
    }
}

void TKSModulator::tksinter_getArgModulationModulationAmount(sSI _nr, YAC_Value *_r) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
		YAC_RETF(arg_modamountmodamount[_nr]);
    }
    else
    {
        YAC_RETF(0.0f);
    }
}

void TKSModulator::tksinter_setArgModulationModulationAmount(sSI _nr, sF32 _amount) {
    if(((sUI)_nr)<MOD_NUMARGS)
    {
		arg_modamountmodamount[_nr]=_amount;
    }
}


void TKSModulator::tksinter_setArray(YAC_Object *_fa) {
    if(YAC_BCHK(_fa, YAC_CLID_FLOATARRAY))
    {
        float_array=(YAC_FloatArray*)_fa;
    }
}

void TKSModulator::tksinter_getArray(YAC_Value *_r) {
    YAC_RETO(float_array, 0);
}

void TKSModulator::tksinter_init(sSI _type, sF32 _amp, sF32 _speed) {
    sU8 i;
    for(i=0; i<MOD_NUMARGS; i++)
    {
        arg_modamount[i]=0.0f;
        arg_modamountmodamount[i]=0.0f;
        arg_mod[i]=0;
        arg_modamountmod[i]=0;
    }
    if( ((sUI)_type)<MOD_NUMTYPES )
    {
        args[MOD_ARG_TYPE]=(sF32)_type;
        args[MOD_ARG_AMP]=_amp;
        args[MOD_ARG_SPEED]=_speed;
    }
}

void TKSModulator::tksinter_tick(sF32 _dt) {
	sF32 spd;
	if(arg_mod[MOD_ARG_SPEED])
	{
		if(arg_modamountmod[MOD_ARG_SPEED])
		{
			sF32 f;
			sF32 r;
			sF32 ma;
			f=arg_modamountmod[MOD_ARG_SPEED]->getValue();
			r=arg_mod[MOD_ARG_SPEED]->getValue();
			ma=(arg_modamountmodamount[MOD_ARG_SPEED]*f)+((1.0f-arg_modamountmodamount[MOD_ARG_SPEED])*arg_modamount[MOD_ARG_SPEED]);
			spd=args[MOD_ARG_SPEED]*(1.0f-ma)+r*ma;
		}
		else
		{
			spd=args[MOD_ARG_SPEED]*(1.0f-arg_modamount[MOD_ARG_SPEED]) + arg_mod[MOD_ARG_SPEED]->getValue()*arg_modamount[MOD_ARG_SPEED];
		}
	}
	else
		spd=args[MOD_ARG_SPEED];//1.0f;

	//tkscript->dcon.printf("spd=%f\n", spd);

    args[MOD_ARG_OFF]+=spd*_dt;

    if(args[MOD_ARG_OFF]>1.0f)
    {
		if((args[MOD_ARG_TYPE]==MOD_NOISE)||(args[MOD_ARG_TYPE]==MOD_NOISE01))
			noiseval=(sF32)(rand()*(1.0f/((sF32)RAND_MAX)));
        args[MOD_ARG_OFF]-=1.0f;
    }
    if(args[MOD_ARG_OFF]<0.0f)
        args[MOD_ARG_OFF]+=1.0f;
    sU8 i;
    for(i=0; i<MOD_NUMARGS; i++)
    {
        if(arg_mod[i])
            arg_mod[i]->tksinter_tick(_dt);
        if(arg_modamountmod[i])
            arg_modamountmod[i]->tksinter_tick(_dt);
    }
}

void TKSModulator::tksinter_getValue(YAC_Value *_r) {
    YAC_RETF(getValue());
}

sF32 TKSModulator::getValue(void) {
    sF32 r;
    sF32 f;
    sF32 ma;
    sF32 av[MOD_NUMARGS];
    sU8 i;
    for(i=0; i<MOD_NUMARGS; i++)
    {
        if(arg_mod[i])
        {
            if(arg_modamountmod[i])
            {
                f=arg_modamountmod[i]->getValue();
                r=arg_mod[i]->getValue();
                ma=(arg_modamountmodamount[i]*f)+((1.0f-arg_modamountmodamount[i])*arg_modamount[i]);
                av[i]=args[i]*(1.0f-ma)+r*ma;
            }
            else
            {
                av[i]=args[i]*(1.0f-arg_modamount[i]) + arg_mod[i]->getValue()*arg_modamount[i];
            }
        }
        else
            av[i]=args[i];//1.0f;
    }
    i=(sU8)av[MOD_ARG_TYPE];
    switch(i)
    {
    default:
        r=0.0f;
        break;
    case MOD_SIN:
        r=(sF32)sin(sM_PI*2.0f*av[MOD_ARG_OFF]);
        break;
    case MOD_TRI:
        if(av[MOD_ARG_OFF]>=0.5f)
            r=(1.0f-av[MOD_ARG_OFF])*4.0f-1.0f;
        else
            r=av[MOD_ARG_OFF]*4.0f-1.0f;
        break;
    case MOD_SAW:
        r=av[MOD_ARG_OFF]*2.0f-1.0f;
        break;
    case MOD_RECT:
        if(av[MOD_ARG_OFF]>=av[MOD_ARG_PULSEWIDTH])
            r=1.0f;
        else
            r=-1.0f;
        break;
    case MOD_NOISE:
        //r=(sF32)(rand()*(2.0f/((sF32)RAND_MAX))-1.0f);
        r=(noiseval*2.0f)-1.0f;
        break;
    case MOD_1BYX:
        r=1.0f-(1.0f/(av[MOD_ARG_OFF]+1.0f)-0.5f)*4.0f;
        break;
    case MOD_ARRAY:
        break;
    case MOD_SIN01:
        r=(sF32)(sin(sM_PI*2.0f*av[MOD_ARG_OFF])+1.0)*0.5f;
        break;
    case MOD_TRI01:
        if(av[MOD_ARG_OFF]>=0.5f)
            r=(1.0f-av[MOD_ARG_OFF])*2.0f;
        else
            r=av[MOD_ARG_OFF]*2.0f;
        break;
    case MOD_SAW01:
        r=av[MOD_ARG_OFF];
        break;
    case MOD_RECT01:
        if(av[MOD_ARG_OFF]>=0.5f)
            r=1.0f;
        else
            r=0.0f;
        break;
    case MOD_NOISE01:
        r=noiseval;
        break;
    case MOD_1BYX01:
        r=1.0f-(1.0f/(av[MOD_ARG_OFF]+1.0f)-0.5f)*2.0f;
        break;
    }

    //YAC_RETF(r);
    return r*av[MOD_ARG_AMP];
}

