///
/// file  : modulatorfactory.cpp
/// author: (c) 2002 Bastian Spiegel <bs@tkscript.de>
///
/// changelog:
///    22042002 <fli> wrote interface, create files
///


#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#define YAC_BIGSTRING defined
#define YAC_PRINTF defined
#include <yac.h>
#include <tks-list.h>
#include "modulator.h"
#include "modulatorfactory.h"

extern sUI class_ID_Modulator;
extern YAC_Host *yac_host;

enum __modfactorystates {
	ST_UNKNOWN=0,
	ST_ARGEXPR,
	ST_VALUE,
	ST_MODEXPR,
	ST_SIN,
	ST_RECT,
	ST_SAW,
	ST_TRI,
	ST_NOISE,
	ST_ARRAY,
	ST_SIN01,
	ST_RECT01,
	ST_SAW01,
	ST_TRI01,
	ST_NOISE01,

	ST_ERROR=0xdeadbeef,

	ST_PAD
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                   ~
//~ TKSModulatorFactory               ~
//~                                   ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
YAC_C(TKSModulatorFactory, "ModulatorFactory");


TKSModulatorFactory::TKSModulatorFactory(void) {
}

TKSModulatorFactory::~TKSModulatorFactory() {
}

#define Dpushstate if(stackindex>4) { state_stack[(--stackindex)]=state; state_stack[(--stackindex)]=state_index; state_stack[--stackindex]=mod_argi; state_stack[--stackindex]=(sUI)mod_pop; } else state=ST_ERROR
#define Dpopstate if(stackindex<2048) { mod_pop=(TKSModulator*)state_stack[stackindex++]; mod_argi=state_stack[stackindex++]; state_index=++state_stack[stackindex++]; state=state_stack[stackindex++]; } else state=ST_ERROR
//#define Dallocnode pid=pool->qAlloc(); if(pid==-1) { yac_host->printf("[---] ModulatorFactory: alloc node failed. maybe your did not allocate a pool that is large enough?\n"); state=ST_ERROR; continue; } else { mod_pop=(TKSModulator*)(*pool)[pid]; if(!mod_root) mod_root=mod_pop; }
#define Dallocnode mod_pop=new TKSModulator(); if(!mod_pop) { yac_host->printf("[---] ModulatorFactory: alloc node failed.\n"); state=ST_ERROR; continue; } else { mod_pop->class_ID=class_ID_Modulator; if(!mod_root) mod_root=mod_pop; }

void TKSModulatorFactory::tksinter_createTree(YAC_Object *_src, YAC_Value *_r) {
	if(YAC_CHK(_src, YAC_CLID_STRING))
	{
		sUI state_stack[2048]; // state_index/state pairs
		sUI stackindex=2048;
		
		YAC_String *src=(YAC_String *)_src;
		sUI soff=0;
		TKSModulator *mod_root=0;
		TKSModulator *mod_pop=0;
		TKSModulator *mod_poparg=0;
		sBool isvalue;
		sF32 float_value;
		sU16 mod_argi=0;
		if(src->chars&&src->length)
		{
			// parse modulator source
			sUI state=ST_UNKNOWN;
			sUI state_index=0;
			Dpushstate;
			sBool allow_value=0;
			YAC_String cw;
			cw.visit((char*)src->chars);
			while(soff<src->length)
			{
				cw.chars=src->chars+soff;
				cw.length=src->length-soff;
				
				isvalue=0;
				switch(cw.chars[0])
				{
				case '\t':
				case '\n':
				case ' ':
				case 0:
					// eat up blanks
					soff++;
					continue;
				}
				
				if(cw.length>1)
				{
					switch(cw.chars[0])
					{
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
						isvalue=1;
						break;
					}
				}
				
				switch(state)
				{
				default:
				case ST_ERROR:
					yac_host->printf("[---] ModulatorFactory: error parsing string.\n");
					//tksinter_disposeTree(mod_root);
					delete mod_root;
					delete mod_poparg;
					YAC_RETO(0, 0);
					return;
					
				case ST_UNKNOWN:
					// any Modulator sequence
					if(state_index>0)
					{
						//mod_poparg=mod_pop;
						Dpopstate;
						break;
					}
					state=ST_MODEXPR; // parse first modulator
					break;
					
				case ST_MODEXPR:
					if(!strncmp(cw.chars, "sin01", 5))
					//if(cw.startsWith("sin01"))
					{
						//Dpushstate;
						state=ST_SIN01;
						state_index=0;
						soff+=5;
						Dallocnode;
						mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_SIN01;
					}
					else 
					{
						if(!strncmp(cw.chars, "rect01", 6))
						//if(cw.startsWith("rect01"))
						{	
							//Dpushstate;
							state=ST_RECT01;
							soff+=6;
							state_index=0;
							Dallocnode;
							mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_RECT01;
						}
						else
						{
							if(!strncmp(cw.chars, "saw01", 5))
							//if(cw.startsWith("saw01"))
							{
								//Dpushstate;
								state=ST_SAW01;
								soff+=5;
								state_index=0;
								Dallocnode;
								mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_SAW01;
							}
							else
							{
								if(!strncmp(cw.chars, "tri01", 5))
								//if(cw.startsWith("tri01"))
								{
									//Dpushstate;
									state=ST_TRI01;
									soff+=5;
									state_index=0;
									Dallocnode;
									mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_TRI01;
								}
								else
								{
									if(!strncmp(cw.chars, "noise01", 7))
									//if(cw.startsWith("noise01"))
									{
										//Dpushstate;
										state=ST_NOISE01;
										soff+=7;
										state_index=0;
										Dallocnode;
										mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_NOISE01;
									}
									else
									{
										if(!strncmp(cw.chars, "sin", 3))
										//if(cw.startsWith("sin"))
										{
											//Dpushstate;
											state=ST_SIN;
											soff+=3;
											state_index=0;
											Dallocnode;
											mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_SIN;
										}
										else
										{
											if(!strncmp(cw.chars, "rect", 4))
											//if(cw.startsWith("rect"))
											{
												//Dpushstate;
												state=ST_RECT;
												soff+=4;
												state_index=0;
												Dallocnode;
												mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_RECT;
											}
											else
											{
												if(!strncmp(cw.chars, "saw", 3))
												//if(cw.startsWith("saw"))
												{
													//Dpushstate;
													state=ST_SIN;
													soff+=3;
													state_index=0;
													Dallocnode;
													mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_SAW;
												}
												else
												{
													if(!strncmp(cw.chars, "tri", 3))
													//if(cw.startsWith("tri"))
													{
														//Dpushstate;
														state=ST_SIN;
														soff+=3;
														state_index=0;
														Dallocnode;
														mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_TRI;
													}
													else
													{
														if(!strncmp(cw.chars, "noise", 5))
														//if(cw.startsWith("noise"))
														{
															//Dpushstate;
															state=ST_NOISE;
															soff+=5;
															state_index=0;
															Dallocnode;
															mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_NOISE;
														}
														else
														{
															if(!strncmp(cw.chars, "array", 5))
															//if(cw.startsWith("array"))
															{
																//Dpushstate;
																state=ST_ARRAY;
																soff+=5;
																state_index=0;
																Dallocnode;
																mod_pop->args[TKSModulator::MOD_ARG_TYPE]=TKSModulator::MOD_ARRAY;
															}
															else
															{
																state=ST_ERROR;
															}
														}
													}
												}
											}
										}
									}
								}
							}
							} 
						} // isvalue
						break;
						
					case ST_ARGEXPR:
						switch(state_index)
						{
						case 0:
							// value expr
							Dpushstate;
							state=ST_VALUE;
							state_index=0;
							break;
						case 1:
							mod_pop->args[mod_argi]=float_value;
							if(cw.chars&&cw.chars[0]==',')
							//if(cw.startsWith(","))
							{
								soff++;
								// modexpr
								Dpushstate;
								state=ST_MODEXPR;
								state_index=0;
							}
							else
							{
								// ST_ARGEXPR finished
								Dpopstate;
							}
							break;
						case 2:
							mod_pop->arg_mod[mod_argi]=mod_poparg;
							if(cw.chars&&cw.chars[0]==',')
							//if(cw.startsWith(","))
							{
								// modamount value
								soff++;
								Dpushstate;
								state=ST_VALUE;
								state_index=0;
							}
							else
							{
								Dpopstate;
							}
							break;
						case 3:
							mod_pop->arg_modamount[mod_argi]=float_value;
							if(cw.chars&&cw.chars[0]==',')
							//if(cw.startsWith(","))
							{
								// modamountmodexpr
								soff++;
								Dpushstate;
								state=ST_MODEXPR;
								state_index=0;
							}
							else
							{
								Dpopstate;
							}
							break;
						case 4:
							mod_pop->arg_modamountmod[mod_argi]=mod_poparg;
							if(cw.chars&&cw.chars[0]==',')
							//if(cw.startsWith(","))
							{
								// modamountmodamount value
								soff++;
								Dpushstate;
								state=ST_VALUE;
								state_index=0;
							}
							else
							{
								Dpopstate;
							}
							break;
						case 5:
							mod_pop->arg_modamountmodamount[mod_argi]=float_value;
							Dpopstate;
							break;
						}
						break;
						
						case ST_VALUE:
							{
								if(!isvalue)
								{
									yac_host->printf("[---] ModulatorFactory::createTree: expected value at s=\"%s\".\n", cw.chars);
									state=ST_ERROR;
								}
								else
								{
									// find end of potential float argument string
									sUI fe_blank;
									sUI fe_nl;
									sUI fe_tab;
									sUI fe_bracket;
									sUI fe_comma;
									sUI fe=(sUI)-2;
									fe_blank=(sUI)cw.indexOf(' ');
									fe_nl=(sUI)cw.indexOf('\n');
									fe_tab=(sUI)cw.indexOf('\t');
									fe_bracket=(sUI)cw.indexOf(')');
									fe_comma=(sUI)cw.indexOf(',');
									if(fe_blank<fe)
										fe=fe_blank;
									if(fe_nl<fe)
										fe=fe_nl;
									if(fe_tab<fe)
										fe=fe_tab;
									if(fe_bracket<fe)
										fe=fe_bracket;
									if(fe_comma<fe)
										fe=fe_comma;
									if(fe>(sUI)-3)
									{
										yac_host->printf("[---] ModulatorFactory::createTree: error parsing float at s=\"%s\".\n", cw.chars);
										state=ST_ERROR;
									}
									else
									{
										YAC_String fs;
										cw.substring(&fs, 0, fe);
										::sscanf(fs.chars, "%f", &float_value);
										//yac_host->printf("fs=\"%s\"\n", fs.chars);
										// xxx parse float
										/*sU16 checkconv=fs.checkConversions();
										switch(checkconv)
										{
										default:
											yac_host->printf("[---] ModulatorFactory::createTree: expected float/int value at s=\"%s\".\n", cw.chars);
											state=ST_ERROR;
											break;
										case 1: // int
											{
												sSI intval=0;
												fs.yacScan(&intval);
												float_value=(sF32)intval;
											}
											break;
										case 2: // float
											{
												fs.yacScan(&float_value);
											}
											break;
										}*/
										soff+=fe;
										allow_value=0;
										Dpopstate;
									}
								}
							}
							break;
							
							
						case ST_RECT:
						case ST_SAW:
						case ST_TRI:
						case ST_NOISE:
						case ST_ARRAY:
						case ST_SIN01:
						case ST_RECT01:
						case ST_SAW01:
						case ST_TRI01:
						case ST_NOISE01:
						case ST_SIN:
							switch(state_index)
							{
							case 0:
								if(cw.chars&&cw.chars[0]=='(')
								//if(cw.startsWith("("))
								{
									soff++;
									state_index++;
								}
								else
								{
									state=ST_ERROR;
								}
								break;
								
							case 1:
								if(!strncmp(cw.chars, "amp=", 4))
								//if(cw.startsWith("amp="))
								{
									soff+=4;
									mod_argi=TKSModulator::MOD_ARG_AMP;
									Dpushstate;
									state=ST_ARGEXPR;
									state_index=0;
									allow_value=0;
								}
								else
								{
									if(!strncmp(cw.chars, "speed=", 6))
									//if(cw.startsWith("speed="))
									{
										soff+=6;
										mod_argi=TKSModulator::MOD_ARG_SPEED;
										Dpushstate;
										state=ST_ARGEXPR;
										state_index=0;
										allow_value=0;
									}
									else
									{
										if(!strncmp(cw.chars, "off=", 4))
										//if(cw.startsWith("off="))
										{
											soff+=4;
											mod_argi=TKSModulator::MOD_ARG_OFF;
											Dpushstate;
											state=ST_ARGEXPR;
											state_index=0;
											allow_value=0;
										}
										else
										{
											if(!strncmp(cw.chars, "pulsewidth=", 11))
											//if(cw.startsWith("pulsewidth="))
											{
												soff+=11;
												mod_argi=TKSModulator::MOD_ARG_PULSEWIDTH;
												Dpushstate;
												state=ST_ARGEXPR;
												state_index=0;
												allow_value=0;
											}
											else
											{
												if(cw.chars&&cw.chars[0]==')')
												//if(cw.startsWith(")"))
												{
													soff++;
													mod_poparg=mod_pop;
													Dpopstate;
													allow_value=0;
													//yac_host->printf("[---] modulator seq finished.\n");
												}
												else
												{
													state=ST_ERROR;
												}
											}
										}
									}
								}
								break;
								
							case 2:
								// ARGEXPR modulator argument finished
								state_index=1;
								//yac_host->printf("[---] modulator argument %i finished.\n",
								//	mod_argi);
								break;
							}
							break;
							
							
							
							
					} // switch state..
				} // while soff<len..
				// return first modulator
				YAC_RETO(mod_root, 1);
				return;
			} // if src->length..
		}
	YAC_RETO(0, 0);
	return;
}


